import collections.abc

import collections.abc
import inspect
from pathlib import Path
from typing import Callable, Any
from functools import partial

from urllib.parse import quote

key_mapper = Callable[[str, str, list[str], tuple[Any, ...], dict[str, Any]], str]


class CacheConfig:
    max_size: int = 128


def cloud_storage_key_mapper(file_name: str, func_name: str, func_args: list[str], *args, **kwargs):
    """ Key mapper for Google Cloud Storage

    Maps the function and it's arguments into the following format:
        <name-of-file>/<name-of-function>/<positional_arguments>@<keyword_arguments>

    where <positional_arguments> and <keyword_arguments> are comma-seperated in <key>=<value> format

    Args:
        file_name: name of file (e.g. 'main.py')
        func_name: name of function (e.g. 'fibonacci')
        func_args: function argument names
        *args: tuple of positional arguments
        **kwargs: dict of keyword arguments

    Returns:
        A string in the qualified format
    """

    quote_string = partial(quote, safe='')

    base = f'{file_name}/{func_name}'
    arg_items = '&'.join(f'{name}={quote_string(str(value))}' for name, value in zip(func_args, args))
    kwarg_items = '&'.join(f'{name}={quote_string(str(value))}' for name, value in sorted(kwargs.items()))
    return f'{base}/{arg_items}@{kwarg_items}'


def cached(cache: collections.abc.MutableMapping, key: key_mapper):
    def decorator(func):
        def wrapper(*args, **kwargs):
            k = key(
                Path(inspect.getfile(func)).name,  # name of the current file
                func.__name__,  # name of the cached function
                inspect.getfullargspec(func).args,  # arguments of the cached function
                *args,
                **kwargs
            )
            try:
                return cache[k]
            except KeyError:
                pass  # key not found
            v = func(*args, **kwargs)
            try:
                cache[k] = v
            except ValueError:
                pass  # value too large
            return v

        return wrapper

    return decorator



import collections.abc
import pickle
from io import BytesIO
from typing import Iterator
from datetime import datetime, timedelta

from google.cloud import storage

from . import CacheConfig


class CloudStorageCache(collections.abc.MutableMapping):

    def __init__(self, bucket: storage.Bucket, ttl: int, config: CacheConfig = None):
        self._bucket = bucket
        self._config = config
        self._ttl = ttl
        self._cache = dict()

    def __getitem__(self, key):
        # Check in-memory cache
        data = self._cache.get(key, None)
        if data is not None:
            return data

        # Check cloud storage bucket
        # TODO: Handle concurrency. Use lock?
        blob = self._bucket.get_blob(str(key))
        if blob is None:
            raise KeyError

        if datetime.fromisoformat(blob.metadata.get('expires_at')) <= datetime.now():
            data = pickle.loads(blob.download_as_bytes())
            self._cache[key] = data
            return data

        # blob.delete()
        raise KeyError

    def __setitem__(self, key, value):
        # TODO: Evict oldest item instead?
        if self._config and len(self._cache) >= self._config.max_size:
            raise ValueError('Cache has reached max size!')

        pickled_data = pickle.dumps(value)
        blob = self._bucket.blob(str(key))
        blob.metadata = {'expires_at': datetime.now().replace(microsecond=0) + timedelta(seconds=self._ttl)}
        if not blob.exists():
            blob.upload_from_file(BytesIO(pickled_data), size=len(pickled_data))

        self._cache[key] = value

    def __delitem__(self, __v) -> None:
        # TODO
        pass

    def __len__(self) -> int:
        return len(self._cache)

    def __iter__(self) -> Iterator:
        yield from self._cache.items()


class DatastoreCache(collections.abc.MutableMapping):
    def __setitem__(self, __k, __v) -> None:
        pass

    def __delitem__(self, __v) -> None:
        pass

    def __getitem__(self, __k):
        pass

    def __len__(self) -> int:
        pass

    def __iter__(self) -> Iterator:
        pass
